/*-
 * Copyright (c) 2012 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>

#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "bigint.h"

int
main(int argc, char **argv)
{
	uint4096_t	stack[32];
	uint4096_t	tmp;
	regmatch_t	match[15];
	uint64_t	bits;
	regex_t		r;
	size_t		nbits;
	char		buf[BUFSIZ];
	int		stackc;
	int		base;
	int		from;
	int		len;
	int		ok;
	int		i;

	nbits = 4096;
	while ((i = getopt(argc, argv, "n:")) != -1) {
		switch(i) {
		case 'n':
			nbits = (size_t)strtoul(optarg, NULL, 0);
			break;
		default:
			break;
		}
	}
	(void) regcomp(&r, "(push[ \t]+((0x)?([0-9a-fA-F]+)))|([-/+*%])|(show)|(pop)|(gcd)|(<<|>>)", REG_EXTENDED);
	stackc = 0;
	for (;;) {
		if (isatty(0)) {
			fprintf(stderr, "fpcalc: ");
		}
		if (fgets(buf, sizeof(buf), stdin) == NULL) {
			break;
		}
		buf[strlen(buf) - 1] = 0x0;
		if (regexec(&r, buf, 15, match, 0) != 0) {
			continue;
		}
		if (match[6].rm_so >= 0) {
			/* show */
			if (stackc > 0) {
				bigint_fprintf(stdout, &stack[stackc - 1]);
				printf("\n");
			}
		} else if (match[7].rm_so >= 0) {
			/* pop */
			if (stackc > 0) {
				stackc -= 1;
			}
		} else if (match[5].rm_so >= 0) {
			/* +-/%* */
			if (stackc < 2) {
				fprintf(stderr, "STACK UNDERFLOW\n");
				continue;
			}
			ok = bigint_op(&tmp, &stack[stackc - 1], buf, &stack[stackc - 2]);
			if (ok) {
				(void) memcpy(&stack[stackc-- - 2], &tmp, sizeof(tmp));
			}
		} else if (match[8].rm_so >= 0) {
			/* gcd */
			if (stackc < 2) {
				fprintf(stderr, "STACK UNDERFLOW\n");
				continue;
			}
			ok = bigint_gcd(&tmp, &stack[stackc - 1], &stack[stackc - 2]);
			if (ok) {
				(void) memcpy(&stack[stackc-- - 2], &tmp, sizeof(tmp));
			}
		} else if (match[9].rm_so >= 0) {
			/* << or >> */
			if (stackc < 2) {
				fprintf(stderr, "STACK UNDERFLOW\n");
				continue;
			}
			if (bigint_uint64(&stack[stackc - 1], &bits)) {
				bigint_word_op(&tmp, &stack[stackc - 2], buf, bits);
				(void) memcpy(&stack[stackc-- - 2], &tmp, sizeof(tmp));
			}
		} else {
			/* got a push */
			if (stackc == __arraycount(stack) - 1) {
				fprintf(stderr, "STACK OVERFLOW\n");
				continue;
			}
			from = 5;
			len = strlen(buf) - from;
			base = 10;
			if (strncmp(&buf[from], "0x", 2) == 0) {
				base = 16;
				from += 2;
				len -= 2;
			}
			bigint_str2bigint(&stack[stackc++], &buf[from], len, base);
		}
	}
	exit(EXIT_SUCCESS);
}
